import type { Oas3Rule, Oas2Rule } from '../../visitors';
import { isNamedType } from '../../types';
import { oasTypeOf, matchesJsonSchemaType, getSuggest } from '../utils';
import { isRef } from '../../ref-utils';

export const OasSpec: Oas3Rule | Oas2Rule = () => {
  return {
    any(node: any, { report, type, location, key, resolve }) {
      const nodeType = oasTypeOf(node);
      if (type.items) {
        if (nodeType !== 'array') {
          report({
            message: `Expected type \`${type.name}\` (array) but got \`${nodeType}\``,
          });
        }
        return;
      } else if (nodeType !== 'object') {
        report({
          message: `Expected type \`${type.name}\` (object) but got \`${nodeType}\``,
        });
        return;
      }

      const required =
        typeof type.required === 'function' ? type.required(node, key) : type.required;
      for (let propName of required || []) {
        if (!(node as object).hasOwnProperty(propName)) {
          report({
            message: `The field \`${propName}\` must be present on this level.`,
            location: [{ reportOnKey: true }],
          });
        }
      }

      for (const propName of Object.keys(node)) {
        const propLocation = location.child([propName]);
        let propValue = node[propName];

        let propType = type.properties[propName];
        if (propType === undefined) propType = type.additionalProperties;
        if (typeof propType === 'function') propType = propType(propValue, propName);

        if (isNamedType(propType)) {
          continue; // do nothing for named schema, it is executed with the next any call
        }

        const propSchema = propType;
        const propValueType = oasTypeOf(propValue);

        if (propSchema === undefined) {
          if (propName.startsWith('x-')) continue;
          report({
            message: `Property \`${propName}\` is not expected here.`,
            suggest: getSuggest(propName, Object.keys(type.properties)),
            location: propLocation.key(),
          });
          continue;
        }

        if (propSchema === null) {
          continue; // just defined, no validation
        }

        if (propSchema.resolvable !== false && isRef(propValue)) {
          propValue = resolve(propValue).node;
        }

        if (propSchema.enum) {
          if (!propSchema.enum.includes(propValue)) {
            report({
              location: propLocation,
              message: `\`${propName}\` can be one of the following only: ${propSchema.enum
                .map((i) => `"${i}"`)
                .join(', ')}.`,
              suggest: getSuggest(propValue, propSchema.enum),
            });
          }
        } else if (propSchema.type && !matchesJsonSchemaType(propValue, propSchema.type)) {
          report({
            message: `Expected type \`${propSchema.type}\` but got \`${propValueType}\`.`,
            location: propLocation,
          });
        } else if (propValueType === 'array' && propSchema.items?.type) {
          const itemsType = propSchema.items?.type;
          for (let i = 0; i < propValue.length; i++) {
            const item = propValue[i];
            if (!matchesJsonSchemaType(item, itemsType)) {
              report({
                message: `Expected type \`${itemsType}\` but got \`${oasTypeOf(item)}\`.`,
                location: propLocation.child([i]),
              });
            }
          }
        }
      }
    },
  };
};
